/************************************************************************
 *   IRC - Internet Relay Chat, m_core/m_knock.c
 *   This file is copyright (C) 2001 Andrew Suffield
 *                                    <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "ircd.h"
#include "common.h"
#include "numeric.h"
#include "irc_string.h"
#include "send.h"
#include "umodes.h"
#include "msg.h"
#include "handlers.h"
#include "channel.h"
#include "hash.h"

#include <string.h>

int m_knock(struct Client *, struct Client *, int, char **);

struct Message knock_msgtab = {
  "KNOCK", 0, 2, 0, MFLG_SLOW, 0,
  {m_unregistered, m_knock, m_knock},
  NULL, NULL
};

/* m_knock
**    parv[0] = sender prefix
**    parv[1] = channel
**  The KNOCK command has the following syntax:
**   :<sender> KNOCK <channel>
**  If a user is not banned from the channel they can use the KNOCK
**  command to have the server NOTICE the channel operators notifying
**  they would like to join.  Helpful if the channel is invite-only, the
**  key is forgotten, or the channel is full (INVITE can bypass each one
**  of these conditions.  Concept by Dianora <db@db.net> and written by
**  <anonymous>
**
** Just some flood control added here, five minute delay between each
** KNOCK -Dianora
**/
int
m_knock(struct Client *cptr, struct Client *sptr, int parc, char *parv[])
{
  struct Channel *chptr;
  char *p, *name;
  struct channel_mask *mask;

  /* anti flooding code,
   * I did have this in parse.c with a table lookup
   * but I think this will be less inefficient doing it in each
   * function that absolutely needs it
   *
   * -Dianora
   */
  static time_t last_used = 0L;

  if (parc < 2)
    {
      sendto_one(sptr, form_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		 "KNOCK");
      return 0;
    }


  /* We will cut at the first comma reached, however we will not *
   * process anything afterwards.                                */

  p = strchr(parv[1], ',');
  if (p)
    *p = '\0';
  name = parv[1];		/* strtoken(&p, parv[1], ","); */

  if (!IsChannelName(name) || !(chptr = hash_find_channel(name, NullChn)))
    {
      sendto_one(sptr, form_str(ERR_NOSUCHCHANNEL), me.name, parv[0], name);
      return 0;
    }

  if (!((chptr->mode_flags & CMODE_INVITEONLY) ||
	(chptr->mode_intval[cmodes[(unsigned char) 'l'].value] &&
	 chptr->users >=
	 chptr->mode_intval[cmodes[(unsigned char) 'l'].value])))
    {
      sendto_one(sptr, get_str(STR_CHANNELOPEN),
		 me.name, sptr->name, chptr->chname);
      return 0;
    }

  mask = build_mask(sptr);
  /* don't allow a knock if the user is banned/quieted, or the channel is secret */
  if ((chptr->mode_flags & CMODE_SECRET) ||
      match_channel_mask(mask, chptr, 'b') ||
      match_channel_mask(mask, chptr, 'q'))
    {
      sendto_one(sptr, form_str(ERR_CANNOTSENDTOCHAN), me.name, parv[0],
		 name);
      return 0;
    }

  /* if the user is already on channel, then a knock is pointless! */
  if (IsMember(sptr, chptr))
    {
      sendto_one(sptr, get_str(STR_ALREADYONCHANNEL),
		 me.name, sptr->name, chptr->chname);
      return 0;
    }

  /* flood control server wide, clients on KNOCK
   * opers are not flood controlled.
   *
   * sure they are, just not godopers -- asuffield
   */

  if (!HasUmode(sptr, UMODE_GOD))
    {
      if ((last_used + PACE_WAIT) > CurrentTime)
	return 0;
      else
	last_used = CurrentTime;
    }

  /* flood control individual clients on KNOCK
   * the ugly possibility still exists, 400 clones could all KNOCK
   * on a channel at once, flooding all the ops. *ugh*
   * Remember when life was simpler?
   * -Dianora
   */

  /* opers are not flow controlled here */
  /* <ahem> *god*opers are not flow controlled -- asuffield */
  if (!HasUmode(sptr, UMODE_GOD)
      && (sptr->last_knock + KNOCK_DELAY) > CurrentTime)
    {
      sendto_one(sptr, get_str(STR_THROTTLEKNOCK),
		 me.name,
		 sptr->name, KNOCK_DELAY - (CurrentTime - sptr->last_knock));
      return 0;
    }

  sptr->last_knock = CurrentTime;

  sendto_one(sptr, get_str(STR_KNOCKED), me.name, sptr->name, chptr->chname);

  /* using &me and me.name won't deliver to clients not on this server
   * so, the notice will have to appear from the "knocker" ick.
   *
   * Ideally, KNOCK would be routable. Also it would be nice to add
   * a new channel mode. Perhaps some day.
   * For now, clients that don't want to see KNOCK requests will have
   * to use client side filtering. 
   *
   * -Dianora
   */

  {
    char message[NICKLEN * 2 + CHANNELLEN + USERLEN + HOSTLEN + 30];

    /* bit of paranoid, be a shame if it cored for this -Dianora */
    if (sptr->user)
      {
	ircsprintf(message, "KNOCK: %s (%s [%s@%s] has asked for an invite)",
		   chptr->chname, sptr->name, sptr->username, sptr->host);
	sendto_channel_type_notice(cptr, chptr, CUFLAG_CHANOP, message);
      }
  }
  return 0;
}
